/*	Experimental TCP over ATM module v0.1
   	This code manipulates ATM virtual circuits in the context of TCP connections.
   	Handles opening, closing, and output capturing.
   	Input capturing is handled by a patch to the ATM vcc input modules.
   	
   	Will M. Miles, t5sq8@unb.ca, Student #251985
   	EE 4013/4023 Undergraduate Thesis, Summer/Fall 2000
*/


/* Version history:
	0.1 - Initial code.
*/


#ifdef TCP_OVER_ATM

int tcpatm_send(struct sk_buff *skb, struct sock *sk)
{	return -1;}

int tcpatm_open(struct sk_buff *skb, struct sock *sk)
/* Performs reverse ATMARP and attempts to open an appropriate connection. 
   A lot of this code was swiped from CLIP, which.. guess what.. has to do the
   exact same thing every time it wants to talk to a new machine. */
{
	struct clip_priv *clip_priv = PRIV(dev);
	struct atmarp_entry *entry;
	struct atm_vcc *vcc;
	int old;
	unsigned long flags;

	DPRINTK("T/A: Open (skb %p)\n",skb);
	if (!skb->dst) {
		/* Abort, no destination. */
		return 0;
	}
	if (!skb->dst->neighbour) {
		/* Abort, not a neighbour. */
		return 0;
	}
	entry = NEIGH2ENTRY(skb->dst->neighbour);
	
	/* I don't understand this.  Clipped a bit off, will leave the rest alone. */
	if (!entry->vccs) {
		if (time_after(jiffies, entry->expires)) {
			/* should be resolved */
			entry->expires = jiffies+ATMARP_RETRY_DELAY*HZ;
			to_atmarpd(act_need,PRIV(dev)->number,entry->ip);
		}
		return 0;
	}
	
	/* There will already be a vcc to our neighbour.  Dupe it and we'll modify it. */
	sk->tp_pinfo.af_tcp.atmlink = vcc = entry->vccs->vcc;
	
	vcc.flags = 0; /* Clear flags. */
	
	/* Set our SAP. */
	memset(vcc.sap,0,sizeof(struct atm_sap));
	vcc.sap.blhi.hl_type = ATM_HL_TCP;  /* Woohoo. */
	
	/* We'll inherit QOS because I'm lazy. */
	
	/* Socket back pointer, weird usage, but.. */
	vcc.sk = sk;
	
	vcc.prev=vcc.next=NULL;

/* Clear stuff, from socket init. */	
	vcc->alloc_tx = alloc_tx;
	vcc->callback = NULL;
	atomic_set(&vcc->tx_inuse,0);
	atomic_set(&vcc->rx_inuse,0);
	vcc->push = NULL;
	vcc->pop = NULL;
	vcc->push_oam = NULL;
	vcc->vpi = vcc->vci = 0; /* no VCI/VPI yet */
	vcc->atm_options = vcc->aal_options = 0;
	vcc->timestamp.tv_sec = vcc->timestamp.tv_usec = 0;
	init_waitqueue_head(&vcc->sleep);
	skb_queue_head_init(&vcc->recvq);
	skb_queue_head_init(&vcc->listenq);
	
	sk->sleep = &vcc->sleep; /* Dangerous...? */

	/* Begin connect procedure.  Ripped from svc_connect.  */
	/* Blocking for now.  On abort it will immediately send TCP packet in the standard way. */
	vcc->reply = WAITING;	
	sigd_enq(vcc,as_connect,NULL,NULL,&vcc->remote);
	add_wait_queue(&vcc->sleep,&wait);
	error = 0;
	while (vcc->reply == WAITING && sigd) {
		set_current_state(TASK_INTERRUPTIBLE);
		schedule();
		if (!signal_pending(current)) continue;
		DPRINTK("*ABORT*\n");
			/*
			 * This is tricky:
			 *   Kernel ---close--> Demon
			 *   Kernel <--close--- Demon
		         * or
			 *   Kernel ---close--> Demon
			 *   Kernel <--error--- Demon
			 * or
			 *   Kernel ---close--> Demon
			 *   Kernel <--okay---- Demon
			 *   Kernel <--close--- Demon
			 */
			sigd_enq(vcc,as_close,NULL,NULL,NULL);
			while (vcc->reply == WAITING && sigd) {
				set_current_state(TASK_UNINTERRUPTIBLE);
				schedule();
			}
			if (!vcc->reply)
				while (!test_bit(ATM_VF_RELEASED,&vcc->flags)
				    && sigd) {
					set_current_state(TASK_UNINTERRUPTIBLE);
					schedule();
				}
			remove_wait_queue(&vcc->sleep,&wait);
			clear_bit(ATM_VF_REGIS,&vcc->flags);
			clear_bit(ATM_VF_RELEASED,&vcc->flags);
			clear_bit(ATM_VF_CLOSE,&vcc->flags);
			    /* we're gone now but may connect later */
			error = -EINTR;
			break;
		}
	remove_wait_queue(&vcc->sleep,&wait);
	sk->tp_pinfo.af_tcp.atmlink = vcc
	if (error) return error;
	if (!sigd) return -EUNATCH;
	if (vcc->reply) return vcc->reply;
	/* Hope this works. */
	
	return 0;
}
	
int tcpatm_listen()
{



}

/*int tcpatm_close()
{	return -1;	}*/


/* Entrypoint function.  Not exactly the best implementation but it works. */
/* Catches the buffer before being sent by the standard IP send function. */
int tcpatm_handle(struct sk_buff *skb, struct sock *sk)
{
	/* Check TCP flags. */
	if ((*(unsigned char*)(skb+13))&0x02) /* is SYN */
	{
		if ((*(unsigned char*)(skb+13))&0x10) /* and ACK */
		{
			if (tcpatm_open(skb,sk)>0) /* Blocks? */
				return tcpatm_send(skb, sk);
			else
			{
				sk->tp_pinfo.af_tcp.useatm = 0;
				return -1;
			}
		}
		else /* Just SYN */
		{
			/*tcpatm_listen(options);*/ /* Hrr, I can open a socket for listening, but how will I know to come back? */
			return -1;
		}
	}

	if (sk->tp_pinfo.af_tcp.atmconnected) /* !?? Fix me. */
		return tcpatm_send(skb, sk);
		/* FIN handling ignored at this point.  Socket close will call destroy ATM connection for us, we'll handle it then. */
		/* Wasteful, but I really don't feel like being finicky. */

	sk->tp_pinfo.af_tcp.useatm = 0; /* We should never get here.  Make sure we don't get back. */
	return -1;
}
		
		
#endif